#include <xkbcommon/xkbcommon.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>
#include <gio/gio.h>
#include "virtual-keyboard-unstable-v1.h"
#include "virtual-keyboard.h"
#include "wayland.h"

struct _DiyaVkbKey
{
    DiyaObject parent;
    uint32_t code;
    gchar *key_caps[3];
    uint32_t syms[3];
    gchar *info;
};

G_DEFINE_FINAL_TYPE(DiyaVkbKey, diya_vkb_key, DIYA_TYPE_OBJECT)

static void diya_vkb_key_dispose(GObject *object)
{
    DiyaVkbKey *self = DIYA_VKB_KEY(object);
    g_debug("diya_vkb_key_dispose: %s", diya_object_to_string(object));
    for (int i = 0; i < 3; i++)
    {
        if (self->key_caps[i])
        {
            free(self->key_caps[i]);
        }
    }
    self->code = 0;
    if (self->info)
    {
        free(self->info);
        self->info = NULL;
    }
    G_OBJECT_CLASS(diya_vkb_key_parent_class)->dispose(object);
}

static void diya_vkb_key_init(DiyaVkbKey *self)
{
    self->code = 0;
    memset(self->key_caps, 0, sizeof(self->key_caps));
    memset(self->syms, 0, sizeof(self->syms));
    self->info = NULL;
}

static const gchar *diya_vkb_key_to_string(DiyaObject *object)
{
    DiyaVkbKey *self = DIYA_VKB_KEY(object);
    if (!self->info && self->code != 0)
    {
        self->info = (char *)malloc(128);
        g_snprintf(self->info, 128,
                   "DiyaVkbKey: %d (0x%.4x: %s, 0x%.4x: %s, 0x%.4x: %s)",
                   self->code,
                   self->syms[0],
                   self->key_caps[0],
                   self->syms[1],
                   self->key_caps[1],
                   self->syms[2],
                   self->key_caps[2]);
    }
    return self->info;
}

static void diya_vkb_key_class_init(DiyaVkbKeyClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    DiyaObjectClass *base_class = DIYA_OBJECT_CLASS(class);

    gobject_class->dispose = diya_vkb_key_dispose;
    // gobject_class->set_property = diya_lock_session_set_property;
    // gobject_class->get_property = diya_lock_session_get_property;
    base_class->to_string = diya_vkb_key_to_string;
}

const gchar *diya_vkb_key_get_keycap(DiyaVkbKey *key, uint32_t shift_level)
{
    if (shift_level > 2)
    {
        return NULL;
    }
    return key->key_caps[shift_level];
}

uint32_t diya_vkb_key_get_code(DiyaVkbKey *key)
{
    return key->code;
}

uint32_t diya_vkb_key_is_alphabet(DiyaVkbKey* self)
{
    return (self->code >= 24 && self->code <= 33) ||
            (self->code >= 38 && self->code <= 46) ||
            (self->code >= 52 && self->code <= 58);
}

/**
 * @brief  DiyaVirtualKeyboard class: lowlevel wayland virtual keyboard handle
 *
 */

static struct zwp_virtual_keyboard_manager_v1 *g_virtual_keyboard_manager = NULL;

struct _DiyaVirtualKeyboard
{
    DiyaShellObject parent;
    struct zwp_virtual_keyboard_v1 *keyboard;
    struct xkb_context *vkb_ctx;
    struct xkb_keymap *vkb_keymap;
};

G_DEFINE_FINAL_TYPE(DiyaVirtualKeyboard, diya_virtual_keyboard, DIYA_TYPE_SHELL_OBJECT);

static void diya_virtual_keyboard_dispose(GObject *object)
{
    DiyaVirtualKeyboard *self = DIYA_VIRTUAL_KEYBOARD(object);
    g_debug("diya_virtual_keyboard_dispose");
    if (self->keyboard)
    {
        zwp_virtual_keyboard_v1_destroy(self->keyboard);
    }
    if (self->vkb_keymap)
    {
        xkb_keymap_unref(self->vkb_keymap);
    }
    if (self->vkb_ctx)
    {
        xkb_context_unref(self->vkb_ctx);
    }
    G_OBJECT_CLASS(diya_virtual_keyboard_parent_class)->dispose(object);
}

static void diya_virtual_keyboard_init(DiyaVirtualKeyboard *self)
{
    self->keyboard = NULL;
    self->vkb_ctx = NULL;
    self->vkb_keymap = NULL;
}

static const gchar *diya_virtual_keyboard_to_string(DiyaObject *object)
{
    (void)object;
    return "Diya virtual keyboard instance";
}

static void diya_virtual_keyboard_class_init(DiyaVirtualKeyboardClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    DiyaObjectClass *base_class = DIYA_OBJECT_CLASS(class);

    gobject_class->dispose = diya_virtual_keyboard_dispose;
    // gobject_class->set_property = diya_lock_session_set_property;
    // gobject_class->get_property = diya_lock_session_get_property;
    base_class->to_string = diya_virtual_keyboard_to_string;
}

/*
static void randname(char *buf)
{
    struct timespec ts;
    clock_gettime(CLOCK_REALTIME, &ts);
    long r = ts.tv_nsec;
    for (int i = 0; i < 6; ++i)
    {
        buf[i] = 'A' + (r & 15) + (r & 16) * 2;
        r >>= 5;
    }
}
*/

static int create_keymap_fd(off_t size)
{
    static const char name[] = "/diya-shell-keymap-XXXXXX";
    const char *base;
    char *path;
    int fd;
    int ret;
    // randname(name + sizeof(name) - 7);
    base = getenv("XDG_RUNTIME_DIR");
    if (!base)
    {
        errno = ENOENT;
        return -1;
    }
    path = malloc(strlen(base) + sizeof(name) + 1);
    if (!path)
        return -1;

    strcpy(path, base);
    strcat(path, name);
    g_debug("Create temp file for keymap: %s", path);

    fd = mkstemp(path);
    if (fd >= 0)
    {
        long flags;
        flags = fcntl(fd, F_GETFD);
        if (flags == -1)
        {
            g_critical("fcntl Unable to F_GETFD: %s", strerror(errno));
            close(fd);
            fd = -1;
        }
        else if (fcntl(fd, F_SETFD, flags | FD_CLOEXEC) == -1)
        {
            g_critical("fcntl Unable to F_SETFD(FD_CLOEXEC): %s", strerror(errno));
            close(fd);
            fd = -1;
        }
        unlink(path);
    }
    else
    {
        g_critical("mkstemp Unable to create temp file %s: %s", path, strerror(errno));
    }
    free(path);

    if (fd < 0)
        return -1;
    do
    {
        ret = ftruncate(fd, size);
    } while (ret < 0 && errno == EINTR);
    if (ret < 0)
    {
        close(fd);
        return -1;
    }

    return fd;
}

DiyaVirtualKeyboard *diya_virtual_keyboard_new(DiyaShell *shell, const gchar *keymap_file)
{
    struct xkb_context *vkb_ctx = NULL;
    struct xkb_keymap *vkb_keymap = NULL;
    gsize len = 0;
    vkb_ctx = xkb_context_new(XKB_CONTEXT_NO_FLAGS);
    if (!vkb_ctx)
    {
        g_critical("diya_virtual_keyboard_new: Failed to create XKB context: %s", strerror(errno));
        return NULL;
    }
    if (keymap_file)
    {
        g_debug("diya_virtual_keyboard_new: Load keymap from file: %s", keymap_file);
        FILE *fp = fopen(keymap_file, "r");
        if (!fp)
        {
            g_critical("diya_virtual_keyboard_new: Failed to open file %s: %s", keymap_file, strerror(errno));
            xkb_context_unref(vkb_ctx);
            return NULL;
        }
        vkb_keymap = xkb_keymap_new_from_file(vkb_ctx, fp, XKB_KEYMAP_FORMAT_TEXT_V1, XKB_KEYMAP_COMPILE_NO_FLAGS);
        fclose(fp);
    }
    else
    {
        GError *error = NULL;
        g_warning("No keymap file specified. Loading default keymap from resource");
        GBytes *bytes = g_resources_lookup_data("/dev/iohub/diya/shell/default.keymap", 0, &error);
        if (error != NULL)
        {
            g_critical("Unable to read keymap file from resource %s", error->message);
            g_error_free(error);
            xkb_context_unref(vkb_ctx);
            return NULL;
        }
        vkb_keymap = xkb_keymap_new_from_string(vkb_ctx, (gchar *)g_bytes_get_data(bytes, &len), XKB_KEYMAP_FORMAT_TEXT_V1, XKB_KEYMAP_COMPILE_NO_FLAGS);
        g_bytes_unref(bytes);
    }

    if (!vkb_keymap)
    {
        g_critical("diya_virtual_keyboard_new: Unable to load keymap: %s", strerror(errno));
        xkb_context_unref(vkb_ctx);
        return NULL;
    }

    char *content = xkb_keymap_get_as_string(vkb_keymap, XKB_KEYMAP_FORMAT_TEXT_V1);
    len = strlen(content);
    if (!content)
    {
        g_critical("diya_virtual_keyboard_new: xkb_keymap_get_as_string: %s", strerror(errno));
        xkb_context_unref(vkb_ctx);
        xkb_keymap_unref(vkb_keymap);
        return NULL;
    }

    int fd = create_keymap_fd(len);
    if (fd == -1)
    {
        g_critical("diya_virtual_keyboard_new: create temp file");
        xkb_context_unref(vkb_ctx);
        xkb_keymap_unref(vkb_keymap);
        return NULL;
    }

    void *ptr = mmap(NULL, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (ptr == (void *)-1)
    {
        g_critical("diya_virtual_keyboard_new: error mmap: %s", strerror(errno));
        xkb_context_unref(vkb_ctx);
        xkb_keymap_unref(vkb_keymap);
        close(fd);
        return NULL;
    }
    // g_debug("Using keymap configuration: \n%s", content);
    strcpy((char *)ptr, content);
    free(content);

    DiyaVirtualKeyboard *vkb = g_object_new(DIYA_TYPE_VIRTUAL_KEYBOARD, "shell", shell, NULL);
    vkb->vkb_ctx = vkb_ctx;
    vkb->vkb_keymap = vkb_keymap;
    DiyaWayland *wayland = diya_shell_get_wayland(shell);
    struct wl_seat *seat = diya_wayland_get_seat(wayland);
    vkb->keyboard = zwp_virtual_keyboard_manager_v1_create_virtual_keyboard(g_virtual_keyboard_manager, seat);
    zwp_virtual_keyboard_v1_keymap(vkb->keyboard, WL_KEYBOARD_KEYMAP_FORMAT_XKB_V1, fd, len);

    return vkb;
}

void diya_virtual_keyboard_register(struct wl_registry *registry, uint32_t name, DiyaShell *shell)
{
    (void)shell;
    g_virtual_keyboard_manager = wl_registry_bind(registry, name, &zwp_virtual_keyboard_manager_v1_interface, 1);
}

void diya_virtual_keyboard_send_key(DiyaVirtualKeyboard *self, uint32_t key, diya_vkb_state_t state)
{
    GDateTime *now = g_date_time_new_now_local();
    uint32_t current_time_ms = g_date_time_get_microsecond(now) / 1000;
    g_date_time_unref(now);
    zwp_virtual_keyboard_v1_key(self->keyboard, current_time_ms, key, (uint32_t)state);
}

gchar *diya_virtual_keyboard_get_keycap(const xkb_keysym_t *sym)
{
    char buf[32];
    switch (*sym)
    {
    case XKB_KEY_Escape:
        g_snprintf(buf, sizeof(buf), "%s", "Esc");
        break;
    case XKB_KEY_BackSpace:
        g_snprintf(buf, sizeof(buf), "%s", "⌫");
        break;
    case XKB_KEY_Tab:
        g_snprintf(buf, sizeof(buf), "%s", "⇥");
        break;
    case XKB_KEY_Return:
        g_snprintf(buf, sizeof(buf), "%s", "⏎");
        break;
    case XKB_KEY_Control_L:
        g_snprintf(buf, sizeof(buf), "%s", "Ctrl");
        break;
    case XKB_KEY_Shift_L:
        g_snprintf(buf, sizeof(buf), "%s", "⇧");
        break;
    case XKB_KEY_Caps_Lock:
        g_snprintf(buf, sizeof(buf), "%s", "⇪");
        break;
    case XKB_KEY_Alt_L:
        g_snprintf(buf, sizeof(buf), "%s", "⎇");
        break;
    case XKB_KEY_Alt_R:
    case XKB_KEY_ISO_Level3_Shift:
        g_snprintf(buf, sizeof(buf), "%s", "⇮");
        break;
    case XKB_KEY_space:
        g_snprintf(buf, sizeof(buf), "%s", "⎵");
        break;
    case XKB_KEY_Up:
        g_snprintf(buf, sizeof(buf), "%s", "↑");
        break;
    case XKB_KEY_Down:
        g_snprintf(buf, sizeof(buf), "%s", "↓");
        break;
    case XKB_KEY_Left:
        g_snprintf(buf, sizeof(buf), "%s", "←");
        break;
    case XKB_KEY_Right:
        g_snprintf(buf, sizeof(buf), "%s", "→");
        break;
    case XKB_KEY_function:
        g_snprintf(buf, sizeof(buf), "%s", "Fn");
        break;
    case XKB_KEY_Meta_L:
        g_snprintf(buf, sizeof(buf), "%s", "⌘");
        break;
    case XKB_KEY_Super_L:
    g_snprintf(buf, sizeof(buf), "%s", "●");
        break;
    case XKB_KEY_dead_circumflex:
        g_snprintf(buf, sizeof(buf), "%s", "^");
        break;
    case XKB_KEY_dead_diaeresis:
        g_snprintf(buf, sizeof(buf), "%s", "¨");
        break;
    default:
        if (xkb_keysym_to_utf8(*sym, buf, sizeof(buf)) == 0)
        {
            return NULL;
        }
        break;
    }
    return g_strdup(buf);
}

DiyaVkbKey *diya_virtual_keyboard_get_key(DiyaVirtualKeyboard *self, uint32_t code)
{
    DiyaVkbKey *key = g_object_new(DIYA_TYPE_VKB_KEY, NULL);
    key->code = code;
    int n_level = xkb_keymap_num_levels_for_key(self->vkb_keymap, code, 0);
    if (n_level > 0)
    {
        n_level = n_level > 3 ? 3 : n_level;
        for (int level = 0; level < n_level; level++)
        {
            const xkb_keysym_t *syms = NULL;
            int n_sym = xkb_keymap_key_get_syms_by_level(self->vkb_keymap, code, 0, level, &syms);
            if (n_sym <= 0 || syms[0] == key->syms[0] || syms[0] == key->syms[1])
            {
                continue;
            }
            key->syms[level] = syms[0];
            key->key_caps[level] = diya_virtual_keyboard_get_keycap(&syms[0]);
        }
    }
    return key;
}